-- Copyright 2017 Stanford University
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.

-- runs-with:
-- [
--  ["-ll:cpu", "4", "-fbounds-checks", "1", "-fdebug", "1",
--   "-fparallelize-dop", "5"],
--  ["-ll:cpu", "4", "-fparallelize-dop", "12"],
--  ["-ll:cpu", "4", "-fparallelize-dop", "4", "-fopenmp", "1"]
-- ]

import "regent"

local c = regentlib.c

fspace fs
{
  f : double,
  g : double,
  h : double,
}

__demand(__parallel)
task init(r : region(ispace(int3d), fs))
where reads writes(r)
do
  for e in r do e.f = 0.3 * (e.x + 1) + 0.3 * (e.y + 1) + 0.4 (e.z + 1) end
  for e in r do e.g = 0 end
  for e in r do e.h = 0 end
end

__demand(__parallel)
task stencil(r : region(ispace(int3d), fs))
where reads(r.f), reads writes(r.g)
do
  var ts_start = c.legion_get_current_time_in_micros()
  __demand(__openmp)
  for e in r do
    e.g = 0.5 * (e.f +
                 r[(e + { 2,  0,  0}) % r.bounds].f +
                 r[(e + { 0,  1,  0}) % r.bounds].f +
                 r[(e + { 1,  2,  0}) % r.bounds].f +
                 r[(e + { 0, -2, -2}) % r.bounds].f +
                 r[(e + { 0,  0, -1}) % r.bounds].f +
                 r[(e + { 0, -1,  0}) % r.bounds].f)
  end
  var ts_end = c.legion_get_current_time_in_micros()
  c.printf("parallel version: %lu us\n", ts_end - ts_start)
end

task stencil_serial(r : region(ispace(int3d), fs))
where reads(r.f), reads writes(r.h)
do
  var ts_start = c.legion_get_current_time_in_micros()
  for e in r do
    e.h = 0.5 * (e.f +
                 r[(e + { 2,  0,  0}) % r.bounds].f +
                 r[(e + { 0,  1,  0}) % r.bounds].f +
                 r[(e + { 1,  2,  0}) % r.bounds].f +
                 r[(e + { 0, -2, -2}) % r.bounds].f +
                 r[(e + { 0,  0, -1}) % r.bounds].f +
                 r[(e + { 0, -1,  0}) % r.bounds].f)
  end
  var ts_end = c.legion_get_current_time_in_micros()
  c.printf("serial version: %lu us\n", ts_end - ts_start)
end

local cmath = terralib.includec("math.h")

task test(size : int)
  c.srand48(12345)
  var is = ispace(int3d, {size, size, size})
  var primary_region = region(is, fs)
  init(primary_region)
  stencil(primary_region)
  stencil_serial(primary_region)
  for e in primary_region do
    regentlib.assert(cmath.fabs(e.h - e.g) < 0.000001, "test failed")
  end
end

task toplevel()
  test(10)
  test(100)
end

regentlib.start(toplevel)
